"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Form = void 0;
const utils = require("../utils");
const components = require("../components");
const field_1 = require("./field");
class Form {
    constructor(name) {
        this.getKeys = () => {
            return this.fields.keys();
        };
        this.valueExists = (path) => {
            const pathLowerCase = path.toLowerCase();
            return utils.exists(this.values, pathLowerCase);
        };
        this.getValue = (path) => {
            const pathLowerCase = path.toLowerCase();
            return utils.get(this.values, pathLowerCase);
        };
        this.setValue = (path, value) => {
            const pathLowerCase = path.toLowerCase();
            if (value !== undefined && value !== "")
                utils.set(this.values, pathLowerCase, value);
            else
                utils.remove(this.values, pathLowerCase);
            this.trigger();
        };
        this.initializeArraysInner = (object, propertyName) => {
            if (typeof object === "object") {
                if (Array.isArray(object)) {
                    const field = this.getField(propertyName);
                    if (field.value === undefined) {
                        field.value = object;
                        field.inputProps.value = object !== null && object !== void 0 ? object : "";
                        field.validate();
                        field.trigger();
                    }
                }
                this.initializeArrays(object, propertyName);
            }
        };
        this.initializeArrays = (object, parentPropertyName = "") => {
            if (Array.isArray(object)) {
                for (let index = 0; index < object.length; index += 1) {
                    const propertyName = parentPropertyName ? `${parentPropertyName}[${index}]` : `[${index}]`;
                    this.initializeArraysInner(object[index], propertyName);
                }
            }
            else {
                for (const key in object) {
                    const propertyName = parentPropertyName ? `${parentPropertyName}.${key}` : key;
                    this.initializeArraysInner(object[key], propertyName);
                }
            }
        };
        this.setValues = (values) => {
            const normalizedValues = utils.lowerCaseKeys(utils.flattenObject(values));
            for (const key in normalizedValues) {
                const field = this.getField(key);
                const value = normalizedValues[key];
                utils.set(this.values, key, value);
                field.value = value;
                field.inputProps.value = value !== null && value !== void 0 ? value : "";
                field.validate();
                field.trigger();
            }
            this.initializeArrays(this.values);
            this.trigger();
        };
        this.setSubmissionErrors = (errors) => {
            const normalizedErrors = utils.lowerCaseKeys(utils.flattenErrorsObject(errors));
            const keys = this.getKeys();
            for (const key of keys) {
                if (key in normalizedErrors) {
                    const fieldErrors = typeof normalizedErrors[key] == "string" ? [normalizedErrors[key]] : normalizedErrors[key];
                    this.getField(key).setSubmissionErrors(fieldErrors);
                }
                else {
                    const field = this.fields.get(key);
                    field.submitFailed = field.validate() === false;
                    field.trigger();
                }
            }
        };
        this.getField = (name) => {
            const nameLowerCase = name.toLowerCase();
            let field = this.fields.get(nameLowerCase);
            if (field === undefined) {
                field = new field_1.Field(this, nameLowerCase, this.getValue(nameLowerCase));
                this.fields.set(nameLowerCase, field);
            }
            return field;
        };
        this.removeField = (name) => {
            const nameLowerCase = name.toLowerCase();
            return this.fields.delete(nameLowerCase);
        };
        this.getValues = () => {
            return Object.assign({}, this.values);
        };
        this.reset = () => {
            const keys = this.getKeys();
            for (const key of keys)
                this.fields.get(key).reset();
        };
        this.checkCanSubmit = () => {
            let allValid = true;
            const keys = this.getKeys();
            for (const key of keys) {
                const field = this.fields.get(key);
                field.submitFailed = field.errors.length > 0;
                if (field.submitFailed) {
                    allValid = false;
                    field.trigger();
                }
            }
            return allValid;
        };
        this.submit = (callback) => {
            if (this.checkCanSubmit()) {
                if (callback)
                    callback(this.values);
                else if (this.onSubmit)
                    this.onSubmit(this.values);
                else
                    console.warn(`Unhandled submit of ${this.name}`, this.values);
                return true;
            }
            else
                return false;
        };
        this.subscribe = (subscriber) => {
            this.subscribers.add(subscriber);
        };
        this.unsubscribe = (subscriber) => {
            this.subscribers.delete(subscriber);
        };
        this.trigger = () => {
            for (const subscriber of this.subscribers)
                subscriber(this.getValues());
        };
        this.name = name !== null && name !== void 0 ? name : utils.uuid();
        this.Form = components.getFormComponent(this);
        this.Submit = components.getSubmitComponent(this);
        this.Reset = components.getResetComponent(this);
        this.Field = components.getFieldComponent(this);
        this.Input = components.getInputComponent(this);
        this.FieldArray = components.getFieldArrayComponent(this);
        this.FieldReader = components.getFieldReaderComponent(this);
        this.onSubmit = undefined;
        this.values = {};
        this.fields = new Map();
        this.subscribers = new Set();
    }
}
exports.Form = Form;
//# sourceMappingURL=form.js.map